# -*- coding: utf-8 -*-
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
#
# DataRobot Python API documentation build configuration file, created by
# sphinx-quickstart on Mon Mar 16 10:50:19 2015.
#
# This file is execfile()d with the current directory set to its containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import shutil
import sys
from zipfile import ZipFile

sys.path.insert(0, os.path.abspath(".."))
from datarobot import __version__ as version  # noqa

# Make sure magic 'tags' field is defined.
# (Not actually useful but makes Cython and pylint errors go away.)
if "tags" not in locals():
    tags = None
if "b" in version:
    # mainline quarterly releases never have the beta "b" in the version, only internal & weekly do
    tags.add("include_experimental_docs")

# Copy the changelog to the docs directory so sphinx will include it
root_dir = os.path.abspath(os.path.dirname(os.path.dirname(__file__)))
changelog_path = os.path.join(root_dir, "CHANGES.rst")
dest_path = os.path.join(root_dir, "docs", "CHANGES.rst")
assert os.path.isfile(changelog_path)
shutil.copyfile(changelog_path, dest_path)

# Create a zip file of the example resources (csv, py, ipynb)
# so that it can be included it as a download within the HTML docs
files_to_zip = []
for root, _, files in os.walk("examples"):
    basename = os.path.basename(root)
    # skip "hidden" directories like notebook autosaves and _build
    if not basename.startswith("_") and not basename.startswith("."):
        for file in files:
            name, ext = os.path.splitext(file)
            ext_lower = ext.lower()
            if ext_lower in [".csv", ".xls", ".xlsx", ".ipynb", ".py"]:
                files_to_zip.append(os.path.join(root, file))
with ZipFile("datarobot_examples.zip", "w") as examples_zip:
    for file in files_to_zip:
        examples_zip.write(file)

# -- General configuration -----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be extensions
# coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "numpydoc",
    "nbsphinx",
]

# make warnings go away- as advised by
# https://github.com/phn/pytpm/issues/3#issuecomment-12133978
numpydoc_show_class_members = False

nbsphinx_execute = "never"

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# The suffix of source filenames.
source_suffix = ".rst"

# Showing automatically documented members in same order as they are in source
autodoc_member_order = "bysource"

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = "index"

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# Thes are both the full X.Y.Z version, including alpha/beta/rc tags.
# We think Sphinx probably wants both to be defined.
release = version

project = u"DataRobot Python Client"
copyright = u"2019, DataRobot, Inc."


# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''
# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ["_build", "**/.ipynb_checkpoints"]

# The reST default role (used for this markup: `text`) to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output ---------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
import sphinx_rtd_theme

html_theme = "sphinx_rtd_theme"
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
# html_theme_options = {}

# Add any paths that contain custom themes here, relative to this directory.
# html_theme_path = []

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
# html_title = None

# A shorter title for the navigation bar.  Default is the same as html_title.
# html_short_title = None

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
# html_logo = None

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
# html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = "DataRobotPythonDoc"


# -- Options for LaTeX output --------------------------------------------------

# latex_elements = {
#     The paper size ('letterpaper' or 'a4paper').
#     'papersize': 'letterpaper',

#     The font size ('10pt', '11pt' or '12pt').
#     'pointsize': '10pt',

#     Additional stuff for the LaTeX preamble.
#     'preamble': '',
# }

fh = open("latex_preamble.tex", "r+")
PREAMBLE = fh.read()
fh.close()
latex_elements = {
    "inputenc": "",
    "utf8extra": "",
    "preamble": PREAMBLE,
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass [howto/manual]).
latex_documents = [
    (
        "index",
        "DataRobotPython_{}_Docs.tex".format(version),
        u"DataRobot Python API Documentation",
        "DataRobot, Inc.",
        "manual",
    ),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output --------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ("index", "datarobotpython", u"DataRobot Python API Documentation", [u"DataRobot, Inc."], 1)
]

# If true, show URL addresses after external links.
# man_show_urls = False


# -- Options for Texinfo output ------------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (
        "index",
        "DataRobotPython",
        u"DataRobot Python API Documentation",
        "DataRobot, Inc.",
        "DataRobotPython",
        "One line description of project.",
        "Miscellaneous",
    ),
]

# Documents to append as an appendix to all manuals.
# texinfo_appendices = []

# If false, no module index is generated.
# texinfo_domain_indices = True

# How to display URL addresses: 'footnote', 'no', or 'inline'.
# texinfo_show_urls = 'footnote'


def setup(app):
    app.add_css_file("css/tight_table.css")
